/*
 * Copyright (c) 2018-2020 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Oct  4 2018
 */
#include "acgps.h"

#include <sys/time.h>

#include <ctype.h>
#include <err.h>
#include <math.h>
#include <poll.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>

#include "codels.h"


static void *		novatel_match(struct gps_channel_s *chan);
static int		novatel_configure(struct gps_device_s *dev);
static int		novatel_unconfigure(struct gps_device_s *dev);
static int		novatel_decode(struct gps_device_s *chan,
                                gps_solution_s *solution,
                                sequence_octet *rtk_data);
static int		novatel_ctl(struct gps_device_s *dev,
                                enum gps_ctl_s request, ...);

const struct gps_driver_s novatel_driver = {
  .protocol = "Novatel",

  .match = novatel_match,
  .configure = novatel_configure,
  .unconfigure = novatel_unconfigure,
  .decode = novatel_decode,
  .ctl = novatel_ctl
};

struct novatel_driver_data_s {
  char model[32];		/* device model */
  int16_t utcoffset;		/* #leap seconds since 6 Jan 1980 */
  double period;		/* configured solution period */

  double tdelay;		/* estimated BEST_XYZ transmission delay */
};


/* --- Novatel protocol ---------------------------------------------------- */

enum novatel_msgid {
  msgid_BESTXYZ = 241,
  msgid_BESTSATS = 1194,
  msgid_DATUM = 160,
  msgid_FIX = 44,
  msgid_LOG = 1,
  msgid_RTCM1 = 107,
  msgid_RTCM3 = 117,
  msgid_RTCM1004 = 770,
  msgid_RTCM1005 = 765,
  msgid_RTCM1006 = 768,
  msgid_RTCM1012 = 891,
  msgid_RTCM1074 = 1475,
  msgid_RTCM1075 = 1476,
  msgid_RTCM1077 = 1478,
  msgid_RTCM1084 = 1482,
  msgid_RTCM1085 = 1483,
  msgid_RTCM1087 = 1485,
  msgid_RTCM1095 = 1490,
  msgid_RTCM1097 = 1492,
  msgid_RTCM1125 = 1596,
  msgid_RTCM1127 = 1598,
  msgid_SERIALCONFIG = 1246,
  msgid_TIME = 101,
  msgid_UNDULATION = 214,
  msgid_UNLOG = 36,
  msgid_UNLOGALL = 38,
  msgid_VERSION = 37
};

enum novatel_portid {
  portid_THISPORT_ALL = 6,
  portid_THISPORT = 192
};

struct __attribute__((packed)) novatel_log {
  uint32_t	port;
  uint16_t	msgid;
  uint8_t	msgtype;
  uint8_t	res[1];
  uint32_t	trigger;
  double	period;
  double	offset;
  uint32_t	hold;
};

struct __attribute__((packed)) novatel_unlog {
  uint32_t	port;
  uint16_t	msgid;
  uint8_t	msgtype;
  uint8_t	res[1];
};

struct __attribute__((packed)) novatel_unlogall {
  uint32_t	port;
  uint32_t	held;
};

struct __attribute__((packed)) novatel_version {
  uint32_t	ncomp;
  struct {
    uint32_t	comptype;
    char	model[16];
    char	psn[16];
    char	hwversion[16];
    char	swversion[16];
    char	bootversion[16];
    char	compdate[12];
    char	comptime[12];
  } comp[];
};

struct __attribute__((packed)) novatel_bestxyz {
  uint32_t	solstat;
  uint32_t	postype;
  double	px, py, pz;
  float		pxsigma, pysigma, pzsigma;
  uint32_t	vsolstat;
  uint32_t	veltype;
  double	vx, vy, vz;
  float		vxsigma, vysigma, vzsigma;
  char		stnid[4];
  float		vlatency;
  float		diffage, solage;
  uint8_t	svs, solnsvs, soll1svs, solnmultisvs;
  uint8_t	res[1];
  uint8_t	extsolstat;
  uint8_t	galbdssigmask, gpsglosigmask;
};

struct __attribute__((packed)) novatel_bestsats {
  uint32_t	entries;
  struct {
    uint32_t	system;
    uint32_t	id;
    uint32_t	status;
    uint32_t	mask;
  } sats[];
};

struct __attribute__((packed)) novatel_datum {
  uint32_t	type;
};

struct __attribute__((packed)) novatel_undulation {
  uint32_t	option;
  float		separation;
};

struct __attribute__((packed)) novatel_fix {
  uint32_t	type;
  double	lat, lon, height;
};

struct __attribute__((packed)) novatel_serialconfig {
  uint32_t	port;
  uint32_t	baud;
  uint32_t	parity, databits, stopbits;
  uint32_t	handshake, brk;
};

struct __attribute__((packed)) novatel_time {
  uint32_t	clockstatus;
  double	offset;
  double	offsetstd;
  double	utcoffset;
  uint32_t	utcyear;
  uint8_t	utcmonth, utcday, utchour, utcmin;
  uint32_t	utcms;
  uint32_t	utcstatus;
};

struct __attribute__((packed)) novatel_response {
  uint32_t	status;
  char		msg[];
};

struct __attribute__((packed)) novatel_hdr {
  uint8_t	sync1, sync2, sync3;
  uint8_t	hlen;
  uint16_t	msgid;
  uint8_t	msgtype;
  uint8_t	port;
  uint16_t	msglen;
  uint16_t	seq;
  uint8_t	idle;
  uint8_t	timestatus;
  uint16_t	week;
  uint32_t	ms;
  uint32_t	status;
  uint16_t	res[1];
  uint16_t	version;
  uint8_t	payload[];
};

struct __attribute__((packed)) novatel_msg {
  struct novatel_hdr hdr;
  union {
    struct novatel_bestxyz bestxyz;
    struct novatel_bestsats bestsats;
    struct novatel_datum datum;
    struct novatel_fix fix;
    struct novatel_log log;
    struct novatel_unlog unlog;
    struct novatel_response response;
    struct novatel_serialconfig serialconfig;
    struct novatel_time time;
    struct novatel_undulation undulation;
    struct novatel_unlogall unlogall;
    struct novatel_version version;
  } data[];
};

#define novatel_msg(len)                                            \
  struct __attribute__((packed)) {                                  \
    struct novatel_msg	msg;                                        \
    uint8_t		payload[len];                               \
    uint32_t		cksum;                                      \
  }

#define novatel_msg_dcl(name, field, type)                              \
  novatel_msg(sizeof(((struct novatel_msg *)0)->data->field)) name = {  \
    .msg.hdr = {                                                        \
      .msgid = type,                                                    \
      .msglen = sizeof(name.payload)                                    \
    }                                                                   \
  };                                                                    \
  name.msg.data->field


static void	novatel_satlist(const struct novatel_bestsats *sats,
                        struct gps_solution_s *solution);
static int	novatel_ack_msg(struct gps_channel_s *chan,
                        struct novatel_msg *msg, int timeout);
static const struct novatel_msg *
		novatel_recv_msg(struct gps_channel_s *chan,
                        const struct timeval *deadline);
static int	novatel_send_msg(struct gps_channel_s *chan,
                        struct novatel_msg *msg);
static uint32_t	novatel_crc32(const struct novatel_hdr *hdr);


/* --- novatel_match ------------------------------------------------------- */

static void *
novatel_match(struct gps_channel_s *chan)
{
  struct novatel_driver_data_s *ddata;
  const struct novatel_msg *msg;
  struct timeval deadline;
  unsigned int i;
  int s;

  /* request version (send an ascii message to detect clones) */
  s = gpstty_sendmsg(chan, "\r\nlog versionb\r\n");
  if (s) return NULL;

  gettimeofday(&deadline, NULL);
  do {
    /* wait for another message for 1s */
    deadline.tv_sec++;
    msg = novatel_recv_msg(chan, &deadline);
    if (!msg) return NULL;

    /* match binary version message */
    if (msg->hdr.msgid != msgid_VERSION) {
      warnx("skipping novatel message %d", msg->hdr.msgid);
      continue;
    }

    /* check OEM version */
    for(i = 0; i < msg->data->version.ncomp; i++)
      if (!strncmp("OEM", msg->data->version.comp[i].hwversion, 3)) break;
    if (i >= msg->data->version.ncomp) /* not compatible */ return NULL;

    /* match */
    ddata = malloc(sizeof(*ddata));
    if (!ddata) return NULL;

    snprintf(ddata->model, sizeof(ddata->model),
             "%.16s", msg->data->version.comp[0].model);
    ddata->utcoffset = 0;
    ddata->period = 0.;
    return ddata;

  } while(1);

  return NULL;
}


/* --- novatel_configure --------------------------------------------------- */

static int
novatel_configure(struct gps_device_s *dev)
{
  struct novatel_driver_data_s *ddata = dev->data;
  const struct novatel_msg *msg;
  struct timeval deadline;
  int s;

  /* delete all logs */
  s = novatel_unconfigure(dev);
  if (s) return s;

  /* UTC offset */
  novatel_msg_dcl(time, log, msgid_LOG) = (struct novatel_log){
    .port = portid_THISPORT,
    .msgid = msgid_TIME,
    .msgtype = 0 /* binary */,
    .trigger = 4 /* once */
  };
  s = novatel_ack_msg(&dev->channel, &time.msg, 1000);
  if (s) return s;

  gettimeofday(&deadline, NULL);
  deadline.tv_sec++;
  while(1) {
    msg = novatel_recv_msg(&dev->channel, &deadline);
    if (!msg) /* no reply */ return 1;

    if (msg->hdr.msgid == msgid_TIME) break;
  }

  ddata->utcoffset = floor(msg->data->time.utcoffset);

  /* WGS84 */
  novatel_msg_dcl(d, datum, msgid_DATUM) = (struct novatel_datum){
    .type = 61 /* WGS84 */
  };
  s = novatel_ack_msg(&dev->channel, &d.msg, 1000);
  if (s) return s;

  novatel_msg_dcl(u, undulation, msgid_UNDULATION) =
    (struct novatel_undulation){
      .option = 1 /* user */,
      .separation = 0.
    };
  s = novatel_ack_msg(&dev->channel, &u.msg, 1000);
  if (s) return s;


  /* periodic messages */
  {
    /* max messages per second for 50% bandwidth */
    double mps = dev->channel.baud / (10. /*8N1*/) / (2. /*50%*/) /
      (sizeof(struct novatel_bestxyz));

    if (mps > 20) ddata->period = 0.05;
    else if (mps > 10) ddata->period = 0.1;
    else if (mps > 5) ddata->period = 0.2;
    else if (mps > 4) ddata->period = 0.25;
    else if (mps > 2) ddata->period = 0.5;
    else if (mps > 1) ddata->period = 1;
    else ddata->period = 1/mps;
  }

  novatel_msg_dcl(data, log, msgid_LOG) = (struct novatel_log){
    .port = portid_THISPORT,
    .msgtype = 0 /* binary */,
    .trigger = 2 /* on time */,
    .period = ddata->period
  };

  data.msg.data->log.msgid = msgid_BESTXYZ;
  s = novatel_ack_msg(&dev->channel, &data.msg, 1000);
  if (s) return s;

  data.msg.data->log.msgid = msgid_BESTSATS;
  data.msg.data->log.period = 10.;
  s = novatel_ack_msg(&dev->channel, &data.msg, 1000);
  if (s) return s;

  ddata->tdelay =
    sizeof(novatel_msg(sizeof(struct novatel_bestxyz))) *
    10./*8N1*//dev->channel.baud;

  return 0;
}


/* --- novatel_unconfigure ------------------------------------------------- */

static int
novatel_unconfigure(struct gps_device_s *dev)
{
  int s;

  /* delete all logs */
  novatel_msg_dcl(unlog, unlogall, msgid_UNLOGALL) = (struct novatel_unlogall){
    .port = portid_THISPORT_ALL,
    .held = 1
  };
  s = novatel_ack_msg(&dev->channel, &unlog.msg, 1000);
  if (s) return s;

  return 0;
}


/* --- novatel_decode ------------------------------------------------------ */

static int
novatel_decode(struct gps_device_s *dev, gps_solution_s *solution,
               sequence_octet *rtk_data)
{
  struct novatel_driver_data_s *ddata = dev->data;
  const struct novatel_msg *msg;
  struct timeval tv;
  int s = EAGAIN;

  while(1) {
    gettimeofday(&tv, NULL);
    msg = novatel_recv_msg(&dev->channel, NULL);
    if (!msg) break;

    switch(msg->hdr.msgid) {
      case msgid_BESTXYZ: {
        const struct novatel_bestxyz *bestxyz = &msg->data->bestxyz;
        or_time_ts utc, ts;

        utc = gpsconv_gpsutc(msg->hdr.week, msg->hdr.ms, ddata->utcoffset);
        ts = gpsconv_utcts(utc, tv, ddata->period, ddata->tdelay);

        solution->ts = ts;
        solution->utc = utc;
        solution->fix = gps_fix_none; /* by default, but updated below */
        solution->sats = bestxyz->solnsvs;

        solution->vel = (gps_enu_s){
          .east = nan(""), .north = nan(""), .up = nan("")
        };
        solution->vel_sigma = (gps_enu_s){
          .east = 0., .north = 0., .up = 0.
        };

        switch(bestxyz->solstat) {
          case 0:
            switch(bestxyz->postype) {
              case  1: solution->fix = gps_fix_fixed; break;
              case 19: solution->fix = gps_fix_interp; break;
              case 16: solution->fix = gps_fix_single; break;
              case 17: solution->fix = gps_fix_diff; break;

              case 32:
              case 33:
              case 34: solution->fix = gps_fix_rtk_float; break;

              case 48:
              case 50: solution->fix = gps_fix_rtk; break;
            }
            break;
        }

        gpsconv_ecefllh(
          &(gps_ecef_s){
            .x = bestxyz->px, .y = bestxyz->py, .z = bestxyz->pz
          }, &solution->llh);
        gpsconv_ecefenu(
          &(gps_ecef_s){
            .x = bestxyz->px, .y = bestxyz->py, .z = bestxyz->pz
          }, &solution->pos);
        gpsconv_ecefenu_stddev(
          &(gps_ecef_s){
            .x = bestxyz->pxsigma, .y = bestxyz->pysigma, .z = bestxyz->pzsigma
          }, &solution->pos_sigma);

        if (bestxyz->vsolstat == 0 && bestxyz->veltype > 0) {
          /* valid velocity */
          gpsconv_decefenu(
            &(gps_ecef_s){
              .x = bestxyz->vx, .y = bestxyz->vy, .z = bestxyz->vz
            }, &solution->vel);
          gpsconv_ecefenu_stddev(
            &(gps_ecef_s){
              .x = bestxyz->vxsigma,
              .y = bestxyz->vysigma,
              .z = bestxyz->vzsigma
            }, &solution->vel_sigma);
        }

        s = 0;
        break;
      }

      case msgid_BESTSATS:
        novatel_satlist(&msg->data->bestsats, solution);
        break;

      case msgid_RTCM1:
      case msgid_RTCM3:
      case msgid_RTCM1004:
      case msgid_RTCM1005:
      case msgid_RTCM1006:
      case msgid_RTCM1012:
      case msgid_RTCM1074:
      case msgid_RTCM1075:
      case msgid_RTCM1077:
      case msgid_RTCM1084:
      case msgid_RTCM1085:
      case msgid_RTCM1087:
      case msgid_RTCM1095:
      case msgid_RTCM1097:
      case msgid_RTCM1125:
      case msgid_RTCM1127: {
        if (!rtk_data) break;

        if (rtk_data->_maximum < rtk_data->_length + msg->hdr.msglen) {
          if (genom_sequence_reserve(
                rtk_data, rtk_data->_length + msg->hdr.msglen))
            return (errno = ENOMEM);
        }
        memcpy(&rtk_data->_buffer[rtk_data->_length],
               msg->hdr.payload, msg->hdr.msglen);
        rtk_data->_length += msg->hdr.msglen;
        s = 0;
        break;
      }

      default:
        warnx("skipping novatel message %d", msg->hdr.msgid);
    }
  }

  return s;
}


/* --- novatel_satlist ----------------------------------------------------- */

static void
novatel_satlist(const struct novatel_bestsats *sats,
                struct gps_solution_s *solution)
{
  static const char *gnssid[] = {
    [0] = "GPS", [1] = "GLONASS", [2] = "SBAS", [5] = "Galileo",
    [6] = "Beidou", [7] = "QZSS"
  };

  sequence8_gps_satlist_s *list = &solution->satlist;
  sequence64_unsigned_short *id;
  gps_satlist_s *slot;
  uint8_t i;
  size_t n;

  /* reset */
  list->_length = 0;
  solution->totalsats = sats->entries;

  /* scan */
  for(i = 0; i < sats->entries; i++) {
    if (sats->sats[i].system >= sizeof(gnssid)/sizeof(*gnssid)) continue;
    if (!gnssid[sats->sats[i].system]) continue;

    /* find gnssid slot or create */
    for(n = 0; n < list->_length; n++)
      if (!strcmp(list->_buffer[n].system, gnssid[sats->sats[i].system]))
        break;
    slot = &list->_buffer[n];

    if (n >= list->_length) {
      if (list->_length >= list->_maximum) continue;
      list->_length++;
      snprintf(slot->system, sizeof(slot->system),
               "%s", gnssid[sats->sats[i].system]);
      slot->used._length = slot->unused._length = 0;
    }

    /* store */
    id = sats->sats[i].mask ? &slot->used : &slot->unused;
    if (id->_length >= id->_maximum) continue;
    id->_buffer[id->_length] = sats->sats[i].id & 0xffff;
    id->_length++;
  }
}


/* --- novatel_ctl --------------------------------------------------------- */

static int
novatel_ctl(struct gps_device_s *dev, enum gps_ctl_s request, ...)
{
  va_list args;
  int s = 0;

  va_start(args, request);
  switch(request) {
    case GPSCTL_BAUD: /* set baud rate */ {
      uint32_t baud = va_arg(args, uint32_t);

      novatel_msg_dcl(c, serialconfig, msgid_SERIALCONFIG) =
        (struct novatel_serialconfig){
        .port = portid_THISPORT_ALL,
        .baud = baud,
        .parity = 0, .databits = 8, .stopbits = 1,
        .handshake = 0, .brk = 1
      };
      s = novatel_ack_msg(&dev->channel, &c.msg, 1000);
      break;
    }

    case GPSCTL_UNFIX: /* unset fixed position */ {
      novatel_msg_dcl(c, fix, msgid_FIX) = (struct novatel_fix){
        .type = 0 /* none */
      };
      s = novatel_ack_msg(&dev->channel, &c.msg, 1000);
      break;
    }

    case GPSCTL_FIX: /* set fixed position */ {
      struct gps_llh_s fix = *va_arg(args, const struct gps_llh_s *);

      novatel_msg_dcl(f, fix, msgid_FIX) = (struct novatel_fix){
        .type = 3 /* position */,
        .lat = fix.latitude, .lon = fix.longitude, .height = fix.height
      };
      s = novatel_ack_msg(&dev->channel, &f.msg, 1000);
      break;
    }

    case GPSCTL_INRTCM: /* feed RTCM data */ {
      uint16_t type = va_arg(args, unsigned int);
      uint8_t *data = va_arg(args, uint8_t *);
      size_t len = va_arg(args, size_t);
      uint16_t msgid;

      if (len == 0) { s = 0; break; }

      switch (type) {
        case 1: msgid = msgid_RTCM1; break;
        case 3: msgid = msgid_RTCM3; break;
        case 1004: msgid = msgid_RTCM1004; break;
        case 1005: msgid = msgid_RTCM1005; break;
        case 1006: msgid = msgid_RTCM1006; break;
        case 1012: msgid = msgid_RTCM1012; break;
        case 1074: msgid = msgid_RTCM1074; break;
        case 1075: msgid = msgid_RTCM1075; break;
        case 1077: msgid = msgid_RTCM1077; break;
        case 1084: msgid = msgid_RTCM1084; break;
        case 1085: msgid = msgid_RTCM1085; break;
        case 1087: msgid = msgid_RTCM1087; break;
        case 1095: msgid = msgid_RTCM1095; break;
        case 1097: msgid = msgid_RTCM1097; break;
        case 1125: msgid = msgid_RTCM1125; break;
        case 1127: msgid = msgid_RTCM1127; break;

        default: s = 1;
      }
      if (s) break;

      novatel_msg(len) rtcm;
      rtcm.msg.hdr.msgid = msgid;
      rtcm.msg.hdr.msglen = len;
      memcpy(rtcm.payload, data, len);

      s = novatel_ack_msg(&dev->channel, &rtcm.msg, 1000);
      break;
    }

    case GPSCTL_OUTRTCM: /* stream RTCM data */ {
      uint16_t rtcmid = va_arg(args, unsigned int);
      double *period = va_arg(args, double *);
      double *offset = va_arg(args, double *);
      uint16_t msgid;

      switch(rtcmid) {
        case 1: msgid = msgid_RTCM1; break;
        case 3: msgid = msgid_RTCM3; break;
        case 1004: msgid = msgid_RTCM1004; break;
        case 1005: msgid = msgid_RTCM1005; break;
        case 1006: msgid = msgid_RTCM1006; break;
        case 1012: msgid = msgid_RTCM1012; break;
        case 1074: msgid = msgid_RTCM1074; break;
        case 1075: msgid = msgid_RTCM1075; break;
        case 1077: msgid = msgid_RTCM1077; break;
        case 1084: msgid = msgid_RTCM1084; break;
        case 1085: msgid = msgid_RTCM1085; break;
        case 1087: msgid = msgid_RTCM1087; break;
        case 1095: msgid = msgid_RTCM1095; break;
        case 1097: msgid = msgid_RTCM1097; break;
        case 1125: msgid = msgid_RTCM1125; break;
        case 1127: msgid = msgid_RTCM1127; break;

        default: return 1;
      }

      if (*period > 0.) {

        if (*period >= 1) *period = round(*period);
        else if (*period >= 0.5) *period = 0.5;
        else if (*period >= 0.25) *period = 0.25;
        else if (*period >= 0.2) *period = 0.2;
        else if (*period >= 0.1) *period = 0.1;
        else *period = 0.05;

        if (*offset >= 1) *offset = round(*offset);
        else if (*offset >= 0.5) *offset = 0.5;
        else if (*offset >= 0.25) *offset = 0.25;
        else if (*offset >= 0.2) *offset = 0.2;
        else if (*offset >= 0.1) *offset = 0.1;
        else *offset = 0;

        /* enable log */
        novatel_msg_dcl(log, log, msgid_LOG) = (struct novatel_log){
          .port = portid_THISPORT,
          .msgid = msgid,
          .msgtype = 0 /* binary */,
          .trigger = 2 /* on time */,
          .period = *period,
          .offset = *offset
        };
        s = novatel_ack_msg(&dev->channel, &log.msg, 1000);

      } else {

        /* cancel log */
        novatel_msg_dcl(unlog, unlog, msgid_UNLOG) = (struct novatel_unlog){
          .port = portid_THISPORT,
          .msgid = msgid,
          .msgtype = 0 /* binary */
        };

        s = novatel_ack_msg(&dev->channel, &unlog.msg, 1000);
      }
      break;
    }

    case GPSCTL_Q_PARAM: /* query model */ {
      struct novatel_driver_data_s *ddata = dev->data;
      gps_param_s *param = va_arg(args, gps_param_s *);

      snprintf(param->model, sizeof(param->model), "%s", ddata->model);
      param->rate = ddata->period > 0. ? 1./ddata->period : 0.;
      break;
    }

    default: s = 1;
  }
  va_end(args);

  return s;
}


/* --- novatel_ack_msg ----------------------------------------------------- */

static int
novatel_ack_msg(struct gps_channel_s *chan, struct novatel_msg *msg,
                int timeout)
{
  const struct novatel_msg *ack;
  struct timeval deadline;
  int s;

  s = novatel_send_msg(chan, msg);
  if (s) return s;

  gettimeofday(&deadline, NULL);
  deadline.tv_usec += timeout * 1000;
  while (deadline.tv_usec > 1000000) {
    deadline.tv_sec++;
    deadline.tv_usec -= 1000000;
  }

  /* ignore all messages until first ACK */
  while(1) {
    ack = novatel_recv_msg(chan, &deadline);
    if (!ack) return -1;

    /* Usually, ack->hdr.msgid == msg->hdr.msgid. However, when sending RTCM
     * messages encapsulated in a binary header, the response has a different
     * message id, that is undocumented. It can even be 0 when the RTCM data in
     * invalid. Since commands are always sent synchronously, it should be
     * safe to ignore the message id in the response and consider the first ACK
     * as the response to the last command. */
    if (ack->hdr.msgtype & 0x80) break;
  }

  if (ack->data->response.status != 1) {
    warnx("command %d: %.*s", msg->hdr.msgid,
          ack->hdr.msglen - 4, ack->data->response.msg);
    return 1;
  }

  return 0;
}


/* --- novatel_recv_msg ---------------------------------------------------- */

static const struct novatel_msg *
novatel_recv_msg(struct gps_channel_s *chan, const struct timeval *deadline)
{
  static const struct novatel_hdr expect = {
    .sync1 = 0xaa, .sync2 = 0x44, .sync3 = 0x12
  };

  const struct novatel_hdr *hdr;
  size_t len, req;
  ssize_t n;

  do {
    /* peek one header */
    hdr = gpstty_data(chan, 0, &len);
    if (len < sizeof(*hdr)) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break; /* err or not enough data */
      continue;
    }

    /* match binary sync bytes */
    if (hdr->sync1 != expect.sync1 ||
        hdr->sync2 != expect.sync2 ||
        hdr->sync3 != expect.sync3) {
      gpstty_data(chan, 1, NULL); /* skip 1 byte */
      continue;
    }

    /* get enough data for the whole message */
    const novatel_msg(hdr->msglen) *msg;

    for(req = sizeof(*msg); len < req; gpstty_data(chan, 0, &len)) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break;
    }

    msg = gpstty_data(chan, 0, &len);
    if (len < req) break; /* err or not enough data */

    /* check crc */
    if (msg->cksum != novatel_crc32(&msg->msg.hdr)) {
      warnx("novatel message %d with bad checksum", msg->msg.hdr.msgid);

      gpstty_data(chan, 3, NULL); /* skip 3 sync bytes */
      continue;
    }

    gpstty_data(chan, req, NULL);
    return &msg->msg;

  } while(1);

  return NULL;
}


/* --- novatel_send_msg ---------------------------------------------------- */

static int
novatel_send_msg(struct gps_channel_s *chan, struct novatel_msg *msg)
{
  msg->hdr = (struct novatel_hdr){
    .sync1 = 0xaa, .sync2 = 0x44, .sync3 = 0x12,
    .hlen = sizeof(struct novatel_hdr),
    .msgid = msg->hdr.msgid,
    .msgtype = 0 /* binary message request */,
    .port = portid_THISPORT,
    .msglen = msg->hdr.msglen,
    .seq = 0,
    .idle = 0,
    .timestatus = 0, .week = 0, .ms = 0
  };
  ((novatel_msg(msg->hdr.msglen) *)msg)->cksum = novatel_crc32(&msg->hdr);

  return gpstty_send(chan, msg, sizeof(novatel_msg(msg->hdr.msglen)));
}


/* --- novatel_crc32 ------------------------------------------------------- */

static uint32_t
novatel_crc32_byte(uint8_t x)
{
  static const uint32_t poly = 0xedb88320l;
  uint32_t crc = x;
  int i;

  for(i = 0; i < 8; i++) {
    if (crc & 1)
      crc = (crc >> 1) ^ poly;
    else
      crc >>= 1;
  }

  return crc;
}

static uint32_t
novatel_crc32(const struct novatel_hdr *hdr)
{
  const uint8_t *p;
  uint32_t crc;
  size_t len;

  len = sizeof(*hdr) + hdr->msglen;
  p = (uint8_t *)hdr;
  crc = 0;

  while(len--)
    crc = (crc >> 8) ^ novatel_crc32_byte(((uint8_t)crc) ^ *p++);

  return crc;
}
