/*
 * Copyright (c) 2018-2021 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Oct  4 2018
 */
#include "acgps.h"

#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>

#include "gps_c_types.h"
#include "codels.h"


/* --- Function set_reference ------------------------------------------- */

/** Codel gps_set_reference of function set_reference.
 *
 * Returns genom_ok.
 * Throws gps_e_sys.
 */
genom_event
gps_set_reference(const gps_llh_s *reference, gps_llh_s *ref,
                  const genom_context self)
{
  (void)self;

  *ref = *reference;
  gpsconv_setref(ref);
  return genom_ok;
}


/* --- Function enable_rtcm --------------------------------------------- */

/** Codel gps_enable_rtcm of function enable_rtcm.
 *
 * Returns genom_ok.
 * Throws gps_e_sys.
 */
genom_event
gps_enable_rtcm(uint16_t rtcmid, sequence_gps_rtcm_recvid *ignored,
                const genom_context self)
{
  (void)self;

  /* remove from the list of ignored incoming messages */
  rtcmidseq_rm(ignored, rtcmid);

  return genom_ok;
}


/* --- Function reset_rtk_stats ----------------------------------------- */

/** Codel reset_rtk_stats of function reset_rtk_stats.
 *
 * Returns genom_ok.
 */
genom_event
reset_rtk_stats(gps_rtcm_stats_s *rtk_stats, const genom_context self)
{
  (void)self;
  size_t i;

  rtk_stats->sent._length = 0;
  rtk_stats->recv._length = 0;
  for (i = 0; i < rtk_stats->ignored._length; i++)
    rtk_stats->ignored._buffer[i].in = 0;

  rtk_stats->inbytes = rtk_stats->inpkts = 0;
  rtk_stats->outbytes = rtk_stats->outpkts = 0;
  rtk_stats->dropbytes = rtk_stats->droppkts = 0;

  return genom_ok;
}


/* --- Function log ----------------------------------------------------- */

/** Codel gps_log of function log.
 *
 * Returns genom_ok.
 * Throws gps_e_sys.
 */
genom_event
gps_log(const char path[64], uint32_t decimation, gps_log_s **log,
        const genom_context self)
{
  int fd;

  fd = open(path, O_WRONLY|O_APPEND|O_CREAT|O_TRUNC, 0666);
  if (fd < 0) return gps_e_sys_error(path, self);

  if (write(fd, gps_log_header_fmt "\n", sizeof(gps_log_header_fmt)) < 0)
    return gps_e_sys_error(path, self);

  if ((*log)->req.aio_fildes >= 0) {
    close((*log)->req.aio_fildes);

    if ((*log)->pending)
      while (aio_error(&(*log)->req) == EINPROGRESS)
        /* empty body */;
  }
  (*log)->req.aio_fildes = fd;
  (*log)->pending = false;
  (*log)->skipped = false;
  (*log)->decimation = decimation < 1 ? 1 : decimation;
  (*log)->missed = 0;
  (*log)->total = 0;

  return genom_ok;
}


/* --- Function log_stop ------------------------------------------------ */

/** Codel gps_log_stop of function log_stop.
 *
 * Returns genom_ok.
 */
genom_event
gps_log_stop(gps_log_s **log, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  if (*log && (*log)->req.aio_fildes >= 0)
    close((*log)->req.aio_fildes);
  (*log)->req.aio_fildes = -1;

  return genom_ok;
}


/* --- Function log_info ------------------------------------------------ */

/** Codel gps_log_info of function log_info.
 *
 * Returns genom_ok.
 */
genom_event
gps_log_info(const gps_log_s *log, uint32_t *miss, uint32_t *total,
             const genom_context self)
{
  (void)self;

  *miss = *total = 0;
  if (log) {
    *miss = log->missed;
    *total = log->total;
  }
  return genom_ok;
}
